

class Dependency(object):
    """Wraps an injected dependency."""

    def __init__(self, name, value, visibility, **kwargs):
        self.name = name
        self.value = value
        self.visibility = visibility





class ConfigurationError(Exception):
    """An :exc:`ConfigurationError` is raised when a malformed or invalid
    configuration is parsed by the framework.
    """
    pass


class RequirementAlreadyResolved(Exception):
    pass


class UnsatisfiedDependency(LookupError):
    """The :exc:`UnsatisfiedDependency` exception is raised when an unknown
    dependency is requested from the provider.
    """
    pass


class EnvironmentVariableNotDefined(EnvironmentError):
    """Raises when an environment variable does not exist."""
    pass


class DependencySatisfied(Exception):
    """Raised when a duplicate dependency is provided."""





class IResolvable(object):
    """Base class for resolvable objects."""

    def has_factory(self):
        return False

    def requires(self):
        return []

    def requires_invocation(self):
        return False

    def is_injected(self):
        return False

    def is_symbol(self):
        return False

    def is_literal(self):
        return False

    def is_collection(self):
        return False

    def resolve(self, resolver):
        return resolver.resolve(self)




import copy
import operator

import six

from ioc.exc import UnsatisfiedDependency

NOT_PROVIDED = object()
NO_DEFAULT = object()


def new_method_proxy(func):
    def inner(self, *args):
        if self._injected is NOT_PROVIDED:
            self._setup()
        value = self._injected
        return func(value, *args)
    return inner


class DeclaredRequirement(object):
    """Represents a requirement declared by a consumer module."""
    _provider = None
    _injected = NOT_PROVIDED
    _names = None

    def __init__(self, provider, names, default=NO_DEFAULT):
        self._injected = NOT_PROVIDED
        self._provider = provider
        self._names = names
        self._default = default
        if isinstance(names, six.string_types):
            self._names = [names]
        self._provider.add_listener(
            self._names,
            lambda k, v: setattr(self, '_injected', NOT_PROVIDED)
        )

    __getattr__ = new_method_proxy(getattr)


    def __setattr__(self, name, value):
        if name in ("_injected","_provider","_names","_default"):
            # Assign to __dict__ to avoid infinite __setattr__ loops.
            self.__dict__[name] = value
        else:
            if self._injected is NOT_PROVIDED:
                self._setup()
            setattr(self._injected, name, value)

    def __delattr__(self, name):
        if name in ("_injected","_provider","_names","_default"):
            raise TypeError("can't delete %s." % name)
        if self._injected is NOT_PROVIDED:
            self._setup()
        delattr(self._injected, name)

    def __call__(self, *args, **kwargs):
        if self._injected is NOT_PROVIDED: # pragma: no cover
            self._setup()
        return self._injected(*args, **kwargs) # pragma: no cover

    def _setup(self):
        try:
            self._injected = self._provider.resolve(self._names[0])
        except UnsatisfiedDependency:
            if self._default is NO_DEFAULT:
                raise
            self._injected = self._default

    def __instancecheck__(self, instance):
        if self._injected is NOT_PROVIDED:
            self._setup()
        return isinstance(instance, type(self._injected))

    def __reduce__(self):
        raise NotImplementedError

    def __getstate__(self):
        raise NotImplementedError

    if six.PY3:
        __bytes__ = new_method_proxy(bytes)
        __str__ = new_method_proxy(str)
        __bool__ = new_method_proxy(bool)
    elif six.PY2:
        __str__ = new_method_proxy(str)
        __unicode__ = new_method_proxy(unicode)
        __nonzero__ = new_method_proxy(bool)

    __copy__ = new_method_proxy(copy.copy)
    __deepcopy__ = new_method_proxy(copy.deepcopy)
    __repr__ = new_method_proxy(repr)
    __int__ = new_method_proxy(int)
    __float__ = new_method_proxy(float)
    __dir__ = new_method_proxy(dir)
    __complex__ = new_method_proxy(complex)
    __round__ = new_method_proxy(round)
    __class__ = property(new_method_proxy(operator.attrgetter("__class__")))
    __eq__ = new_method_proxy(operator.eq)
    __ne__ = new_method_proxy(operator.ne)
    __hash__ = new_method_proxy(hash)
    __getitem__ = new_method_proxy(operator.getitem)
    __setitem__ = new_method_proxy(operator.setitem)
    __delitem__ = new_method_proxy(operator.delitem)
    __iter__ = new_method_proxy(iter)
    __len__ = new_method_proxy(len)
    __contains__ = new_method_proxy(operator.contains)



import copy
import importlib
import itertools

from marshmallow.exceptions import ValidationError
import six
import yaml

from ioc.schema.dependency import Dependency
from ioc.schema.exc import InvalidDeclaration
from ioc.schema.adapters import SymbolDependencyAdapter
from ioc.schema.adapters import LiteralDependencyAdapter
from ioc.schema.adapters import NestedDependencyAdapter
from ioc.schema.adapters import DependencyCollectionAdapter
from ioc.schema.base import Schema


class SchemaParser(object):
    """Parses the dependency configuration schema."""

    def __init__(self, provider=None, override=False):
        self.provider = provider or importlib.import_module('ioc.provider')
        self.override = override

    def load(self, dependencies):
        """Inspect the dependency configuration and validate the declared
        items.
        """
        schema = Schema(self.provider, override=self.override)

        # Check that each dependency has at least a name.
        names = set()
        for dep in dependencies:
            name = dep.get('name')
            if name is None:
                raise InvalidDeclaration(dep, "No name specified")

            if not isinstance(name, six.string_types):
                raise InvalidDeclaration(dep, "Invalid name specified: %s" % repr(name))
            if name in names:
                raise InvalidDeclaration(dep, "Duplicate name: %s" % name)

            names.add(name)

        # For each dependency, try to load them as either a Simple, Literal
        # or NestedDependency.
        adapters = [
            NestedDependencyAdapter(),
            SymbolDependencyAdapter(),
            LiteralDependencyAdapter(),
            DependencyCollectionAdapter(),
        ]

        for dep in dependencies:
            cleaned_dep = None
            for adapter in adapters:
                try:
                    cleaned_dep = adapter.load(copy.deepcopy(dep))
                    break
                except ValidationError as e:
                    print(adapter, e)
                    pass

                cleaned_dep = None

            # If there is still no dependency loaded at this point, this
            # means the declaration was invalid.
            if cleaned_dep is None:
                raise InvalidDeclaration(dep, "Invalid dependency declaration")

            assert issubclass(type(cleaned_dep), Dependency),\
                "load() must return a Dependency implementation, got %s" %\
                repr(cleaned_dep)
            schema.add(cleaned_dep)

        return schema



"""Declares functions to import dependencies."""
import importlib


BUILTINS = __builtins__


def import_module(qualname, fail_silent=False):
    """Import a Python module using its qualified name `qualname`."""
    try:
        return importlib.import_module(qualname)
    except ImportError:
        if not fail_silent:
            raise

        return None


def import_builtin(qualname, fail_silent=False):
    """Import a builtin type."""
    try:
        return BUILTINS[qualname]
    except KeyError:
        if not fail_silent:
            raise ImportError(qualname)

        return None


def import_symbol(qualname, fail_silent=False):
    """Import a symbol. The symbol may be one of the following and
    are considered in this order:

    -   A module;
    -   A class, function or variable in a module;
    -   A builtin type, or a member of a builtin type.
    """
    parts = qualname.rsplit('.', 1)
    if len(parts) == 1:
        result = import_module(parts[0], True) or import_builtin(parts[0], True)
        if result is None and not fail_silent:
            raise ImportError(qualname)

        return result

    namespace, name = parts
    if namespace.count('.') == 0: # Module or builtin type
        parent = import_symbol(namespace, True)
    else:
        parent = import_module(namespace, True)\
            or import_symbol(namespace, True)

    if parent is None and not fail_silent:
        raise ImportError(qualname)

    try:
        return getattr(parent, name)
    except AttributeError:
        if not fail_silent:
            raise ImportError(qualname)

        return None



import keyword
import re

import six


def bytesequence(value, *args, **kwargs):
    return six.binary_type(value)\
        if six.PY2\
        else six.binary_type(value, *args, **kwargs)


def is_valid_identifier(name):
    """Return a boolean indicating if the given `name` is a
    valid path to a Python symbol.
    """
    is_valid = True
    for p in name.split('.'):
        is_valid &= not keyword.iskeyword(p)

        is_valid &= p.isidentifier() if six.PY3\
            else (re.match(r"^[^\d\W]\w*\Z", p) is not None)

    return is_valid



import marshmallow
import six

from ioc import loader
from ioc.schema.requirement import SchemaRequirement
from ioc.schema.dependency import LiteralDependency
from ioc.schema.dependency import SimpleDependency
from ioc.schema.dependency import NestedDependency
from ioc.schema.dependency import DependencyCollection
from ioc.utils import is_valid_identifier


class BaseDependencyAdapter(marshmallow.Schema):

    @marshmallow.post_load
    def initialize_dependency(self, params, **kwargs):
        return SchemaRequirement(**params)

    @marshmallow.validates_schema
    def validate_value(self, params, **kwargs):
        if params.get('type') not in ('symbol','ioc'):
            return

        if not isinstance(params.get('value'), six.string_types):
            raise marshmallow.ValidationError(
                "'value' must hold a valid name",'value')

        if params.get('type') == 'symbol'\
        and not is_valid_identifier(params.get('value')):
            raise marshmallow.ValidationError(
                "'value' must hold a valid name",'value')


class ArgumentDependencyAdapter(BaseDependencyAdapter):
    type = marshmallow.fields.String(
        validate=marshmallow.validate.OneOf(['symbol','ioc','literal']),
        required=True)
    value = marshmallow.fields.Field(required=True)


class CallableDependencyAdapter(BaseDependencyAdapter):
    type = marshmallow.fields.String(
        validate=marshmallow.validate.OneOf(['symbol','ioc']),
        required=True)
    value = marshmallow.fields.Field(required=True)


class SymbolDependencyAdapter(marshmallow.Schema):
    type = marshmallow.fields.String(validate=marshmallow.validate.OneOf(['symbol']),
        required=True)
    name = marshmallow.fields.String(required=True)
    value = marshmallow.fields.String(required=True)
    visibility = marshmallow.fields.String(missing='public', choices=['public','internal'])
    invoke = marshmallow.fields.Boolean(data_key='callable', missing=False)
    args = marshmallow.fields.List(marshmallow.fields.Field, missing=list)
    kwargs = marshmallow.fields.Dict(
        keys=marshmallow.fields.String,
        values=marshmallow.fields.Field,
        missing=dict
    )
    tags = marshmallow.fields.List(
        marshmallow.fields.String,
        required=False,
        missing=list
    )
    mode = marshmallow.fields.String(
        required=False,
        missing='declare',
        validate=[marshmallow.validate.OneOf(['append'])]
    )

    @marshmallow.post_load
    def resolve_symbol(self, params, **kwargs):
        symbol = params['value']
        if symbol == 'bytes':
            params['value'] = 'ioc.utils.bytesequence'
        return SimpleDependency(**params)


class LiteralDependencyAdapter(marshmallow.Schema):
    type = marshmallow.fields.String(
        validate=marshmallow.validate.OneOf(['literal']),
        required=False,
        missing='literal')
    name = marshmallow.fields.String(required=True)
    value = marshmallow.fields.Field(required=True)
    visibility = marshmallow.fields.String(missing='public', choices=['public','internal'])
    tags = marshmallow.fields.List(
        marshmallow.fields.String,
        required=False,
        missing=list
    )

    @marshmallow.post_load
    def resolve(self, params, **kwargs):
        return LiteralDependency(**params)


class NestedDependencyAdapter(marshmallow.Schema):
    name = marshmallow.fields.String(required=True)
    visibility = marshmallow.fields.String(missing='public', choices=['public','internal'])
    factory = marshmallow.fields.Nested(CallableDependencyAdapter, required=True)
    args = marshmallow.fields.List(
        marshmallow.fields.Nested(ArgumentDependencyAdapter),
        required=False,
        missing=list
    )
    kwargs = marshmallow.fields.Dict(
        keys=marshmallow.fields.String,
        values=marshmallow.fields.Nested(ArgumentDependencyAdapter),
        required=False,
        missing=dict
    )
    chain = marshmallow.fields.List(
        marshmallow.fields.Nested(ArgumentDependencyAdapter),
        required=False,
        missing=list
    )
    tags = marshmallow.fields.List(
        marshmallow.fields.String,
        required=False,
        missing=list
    )

    @marshmallow.post_load
    def resolve(self, params, **kwargs):
        if params.get('kwargs'):
            params['kwargs'] = {x: y
                for x,y in params.pop('kwargs').items()}

        return NestedDependency(**params)


class DependencyCollectionAdapter(marshmallow.Schema):
    name = marshmallow.fields.String(required=True)
    members = marshmallow.fields.List(
        marshmallow.fields.String,
        required=True
    )

    @marshmallow.post_load
    def resolve(self, params, **kwargs):
        return DependencyCollection(**params)




import glob
import inspect
import itertools
import functools

import yaml

from ioc import provider
from ioc import schema
from ioc.requirement import DeclaredRequirement
from ioc.requirement import NOT_PROVIDED
from ioc.requirement import NO_DEFAULT
from ioc.exc import UnsatisfiedDependency


__version__ = '1.3.11'


def require(names, *args, **kwargs):
    return DeclaredRequirement(provider, names, *args, **kwargs)


def provide(name, value, force=False, tags=None):
    """Register a Python object as a dependency under the key `name`."""
    provider.register(name, value, force=force, tags=tags)
    return value


def retire(name):
    return provider.retire(name)


def override(name, value):
    return provide(name, value, force=True)


def teardown():
    """Tear down the configured dependencies."""
    provider.teardown()


def load(dependencies, override=False):
    s = schema.Parser(override=override).load(dependencies)
    s.resolve(schema.Resolver())


def load_config(filenames, override=False):
    for filepath in itertools.chain(*[glob.glob(x) for x in filenames]):
        with open(filepath) as f:
            load(yaml.safe_load(f.read()), override=override)


def is_satisfied(name):
    return provider.is_satisfied(name)


def tagged(tag):
    return provider.tagged(tag)


class class_property(object):

    def __init__(self, name, factory=None, default=NO_DEFAULT):
        self.name = name
        self.factory = factory or (lambda x: x)
        self.dep = require(name, default=default)
        self.default = default

    def __get__(self, obj, objtype):
        if self.dep._injected is NOT_PROVIDED:
            self.dep._setup()
        return self.factory(self.dep)


class inject:
    """Ensures that `injected` is made available under
    `name`.

    For decorated classes, this implies that the decorator
    sets an attribute `name` on the class with a
    :class:`class_property` instance pointing to `injected`.
    """

    def __init__(self, name, injected):
        self.attname = name
        self.injected = injected

    def __call__(self, obj):
        if inspect.isclass(obj):
            d = self._decorate_class(obj)
        elif inspect.isfunction(obj):
            d = self._decorate_function(obj)
        else:
            raise ValueError("Can only decorate classes and functions atm.")
        return d

    def _decorate_class(self, obj):
        setattr(obj, self.attname, class_property(self.injected))
        return obj

    def _decorate_function(self, obj):
        @functools.wraps(obj)
        def f(*args, **kwargs):
            kwargs[self.attname] = require(self.injected)
            return obj(*args, **kwargs)
        return f


def call(name, *args, **kwargs):
    """Invoke the dependency identified by `name` with
    the given positional and keyword arguments.
    """
    pass



from ioc.schema.requirement import SchemaRequirement
from ioc.schema.iresolvable import IResolvable


class Dependency(IResolvable):
    """Represents a dependency declared in the configuration
    file.
    """

    @property
    def key(self):
        return self.name

    def __init__(self, visibility='public', tags=None, mode=None, **kwargs):
        self.visibility = visibility
        self.tags = tags or []
        self.mode = mode or None
        self.append_existing = self.mode == 'declare'

    def requires(self):
        """Return a list containing all the requirements of this
        dependency. Only applicable to ``NestedDependency`` instances.
        """
        raise NotImplementedError


class SimpleDependency(Dependency):

    def __init__(self, name, value, invoke=False, args=None, kwargs=None, **params):
        self.name = name
        self.value = value
        self.invoke = invoke or bool(args or kwargs)
        self.args = args or []
        self.kwargs = kwargs or {}
        Dependency.__init__(self, **params)

    def requires_invocation(self):
        return self.invoke

    def is_symbol(self):
        return True


class LiteralDependency(Dependency):

    def __init__(self, name, value, **kwargs):
        Dependency.__init__(self, **kwargs)
        self.name = name
        self.value = value

    def is_literal(self):
        return True


class NestedDependency(Dependency):

    def __init__(self, name, factory, args=None, kwargs=None, chain=None, **params):
        self.name = name
        self.factory = factory
        self.args = args or []
        self.kwargs = kwargs or {}
        self.chain = chain or []
        Dependency.__init__(self, **params)

    def has_factory(self):
        return True

    def is_injected(self):
        return True

    def requires(self):
        return ([self.factory.value] if self.factory.is_injected() else [])\
            + [x.value for x in self.args if x.is_injected()]\
            + [x.value for x in self.kwargs.values() if x.is_injected()]\
            + [x.value for x in self.chain if x.is_injected()]


class DependencyCollection(Dependency):

    def __init__(self, name, members, **kwargs):
        Dependency.__init__(self, **kwargs)
        self.name = name
        self.members = members

    def is_collection(self):
        return True



import six


class DependencyCollection(object):

    def __init__(self, provider, members):
        self._names = members
        self._provider = provider
        self._members = []

    def __getitem__(self, arg):
        members = self._names.__getitem__(arg)
        return [self._provider.resolve(x) for x in members]\
            if isinstance(arg, slice)\
            else self._provider.resolve(members)

    def __iter__(self):
        for i, name in enumerate(self._names):
            yield self._provider.resolve(name)



from collections import defaultdict

from ioc.exc import UnsatisfiedDependency
from ioc.exc import DependencySatisfied
from ioc.requirement import DeclaredRequirement


class Provider(object):

    def __init__(self):
        self.__dependencies = {}
        self.__requirements = {}
        self.__resolved = {}
        self.__listeners = defaultdict(list)
        self.__tags = defaultdict(set)

    def add_listener(self, names, callback):
        """Adds a dependency as a listener to the provider. This
        is used to notify them in changes of dependencies.
        """
        for name in names:
            self.__listeners[name].append(callback)

    def is_satisfied(self, name):
        """Return a boolean indicating if the dependency identified by
        `name` is satisfied.
        """
        return name in self.__resolved

    def resolve(self, name):
        """Return the Python object to which the dependency `name` resolved."""
        if name not in self.__resolved:
            raise UnsatisfiedDependency(name)
        return self.__resolved[name]

    def retire(self, name):
        """Remove a depedency from the registry. This is mainly used
        during testing. Return a boolean indicating if a dependency
        was removed.
        """
        if name not in self.__resolved:
            return False

        del self.__resolved[name]
        for tag in dict.keys(self.__tags):
            if name not in self.__tags[tag]:
                continue
            self.__tags[tag].remove(name)

        return True

    def register(self, name, value, force=False, tags=None):
        """Register `value` as resolved under `name`."""
        if name in self.__resolved and not force:
            raise DependencySatisfied(name)
        self.__resolved[name] = value
        for callback in self.__listeners[name]:
            callback(name, value)

        for tag in (tags or []):
            self.__tags[tag].add(name)

    def provide(self, dep, force=False):
        """Satisfy the requirement identified by :attr:`Dependency.name`,
        optionally force overwrite by providing ``force=True``.
        """
        if force:
            raise NotImplementedError

        if dep.name in self.__dependencies:
            raise DependencySatisfied("Dependency already satisfied: %s" % dep.name)

        self.__dependencies[dep.name] = dep

    def get(self, name, *names):
        """Return the dependency that is identified by `name`."""
        if names:
            raise NotImplementedError

        if name not in self.__dependencies:
            raise UnsatisfiedDependency(name)

        return self.__dependencies.get(name)

    def teardown(self):
        """Tears down all dependencies and requirements."""
        self.__dependencies = {}
        self.__requirements = {}
        self.__resolved = {}

    def tagged(self, tag):
        """Return the identifiers of all dependencies tagged with
        `tag`.
        """
        return list(map(self.resolve, sorted(self.__tags[tag])))

    #def require(self, names):
    #    return DeclaredRequirement(self, names)


provider = Provider()
del Provider

provide = provider.provide
retire = provider.retire
teardown = provider.teardown
get = provider.get
register = provider.register
is_satisfied = provider.is_satisfied
resolve = provider.resolve
add_listener = provider.add_listener
tagged = provider.tagged
#require = provider.require





class Schema(object):
    """Represents a syntactically correct dependency
    configuration.
    """

    def __init__(self, provider, override=False):
        self.provider = provider
        self.override = override
        self.dependencies = []

    def add(self, dep):
        self.dependencies.append(dep)

    def resolve(self, resolver):
        """Resolve all dependencies and register them with the provider. This
        method is atomic: if one dependency fails, none of the dependencies
        are added to the provider.
        """
        resolved = []
        for dep in self.dependencies:
            if dep.mode == 'append':
                seq = self.provider.resolve(dep.name)
                seq.append(dep.resolve(resolver))
            else:
                self.provider.register(
                    dep.key, dep.resolve(resolver), force=self.override,
                    tags=dep.tags
                )



import collections
import contextlib
import inspect


class MissingDependency(LookupError):

    def __init__(self, name, func):
        self.name = name
        self.func = func

    def __str__(self):
        return "%s specifies dependency `%s` but its not provided by "\
            "the current context." % (repr(self.func), self.name)


class ArgumentDependencyInjector:
    """Resolves dependencies by inspecting the argument names
    in a function signature.
    """
    missing = object()
    MissingDependency = MissingDependency

    def __init__(self):
        self.reset()

    def call(self, func):
        """Call `func` and inject arguments based on its signature.

        Args:
            func (callable): the callable that specified dependencies
                in its signature.
        """
        args, kwargs = self.resolve(func)
        return func(*args, **kwargs)

    def reset(self):
        """Unregister all dependencies."""
        self.ctx = {'resolver': self}

    def provide(self, name, value):
        """Provide dependency `name` with `value`."""
        if name == 'injector':
            raise ValueError('`injector` is a reserved name.')
        self.ctx[name] = value

    def resolve(self, func):
        """Inspect the signature of callable `func` and return
        the positional arguments it requires based on the context.

        Args:
            func (callable): the callable that specified dependencies
                in its signature.

        Raises:
            MissingDependency: no dependency was registered for `name`.
        """
        args, varargs, keywords, defaults = inspect.getargspec(func)
        signature = inspect.signature(func)

        # For instance methods, ignore the self argument.
        if inspect.ismethod(func):
            args = args[1:]

        injected_args, injected_kwargs = [], {}
        for arg in args:
            default = signature.parameters[arg].default
            value = self.get(arg)
            if value == self.missing:
                if default == inspect.Parameter.empty:
                    raise self.MissingDependency(arg, func)
                value = default
            injected_args.append(value)

        return tuple(injected_args), injected_kwargs

    def get(self, name):
        """Return the dependency identified by name.

        Args:
            name (str): the name of the dependency.

        Returns:
            object
        """
        return self.ctx.get(name, self.missing)

    def add(self, resolvables):
        """Add one or many resolvables to the resolver."""
        if not isinstance(resolvables, collections.Sequence):
            resolvables = [resolvables]
        assert all([self.is_resolvable(x) for x in resolvables]),\
            repr(resolvables)
        for obj in resolvables:
            self.provide(obj.resolver_name, obj)

    def is_resolvable(self, obj):
        """Return a boolean indicating if `obj` is resolvable."""
        return isinstance(obj, Resolvable)

    def is_provided(self, name):
        """Return a boolean indicating if the dependency identified by
        string `name` is provided.
        """
        return name in self.ctx

    @contextlib.contextmanager
    def context(self, ctx):
        """Setup the context using `ctx` and tear it down after
        exiting.
        """
        for key, value in ctx.items():
            self.provide(key, value)
        try:
            yield
        finally:
            self.reset()


class Resolvable:
    """Mixin class that allows the resolver to detect an instance
    as resolvable and register it.
    """
    resolver_name = None



#!/usr/bin/env python3
#!/usr/bin/env python3
#
# Copyright (C) 2019-2020 Cochise Ruhulessin
#
# This file is part of python-ioc.
#
# python-ioc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# python-ioc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with python-ioc.  If not, see <https://www.gnu.org/licenses/>.
from setuptools import setup


version = str.strip(open('ioc/VERSION').read())
requirements = str.split(open('requirements.txt').read(), '\n')


setup(
    name='python-ioc',
    version=version,
    description='Python Inversion of Control Framework',
    author='Cochise Ruhulessin',
    author_email='cochiseruhulessin@gmail.com',
    url='https://www.wizardsofindustry.net',
    project_name='Inversion of Control',
    install_requires=requirements,
    packages=[
        'ioc',
        'ioc.schema',
    ]
)



from ioc.schema.iresolvable import IResolvable


class SchemaRequirement(IResolvable):
    """Represents a requirement stated in the dependency configuration
    file.
    """

    @property
    def key(self):
        if not self.is_injected():
            raise AttributeError('key')
        return self.value

    def __init__(self, type, value, **params):
        self.type = type
        self.value = value
        self.mode = params.get('mode') or 'declare'

    def is_symbol(self):
        return self.type == 'symbol'

    def is_injected(self):
        return self.type == 'ioc'

    def is_literal(self):
        return self.type == 'literal'

    def resolve(self, resolver):
        """Resolve the dependency using `resolver`."""
        return resolver.resolve(self)

    def __iter__(self):
        return iter([self.type, self.value])

    def __hash__(self):
        return hash(type(self).__name__ + self.type + self.value)

    def __eq__(self, other):
        return hash(self) == hash(other) and type(self) == type(other)



import os

from ioc.loader import import_symbol
from ioc.exc import ConfigurationError
from ioc.exc import UnsatisfiedDependency
from ioc.exc import EnvironmentVariableNotDefined
from ioc import provider
from ioc.collection import DependencyCollection


class SchemaResolver(object):
    """Resolves the dependencies declared in a schema."""

    def resolve(self, dep):
        """Resolves a dependency `dep` and returns its declared Python
        object.
        """

        # If a string value on a dependency starts with a dollar-sign, it is
        # considered to be an environment variable.
        if isinstance(getattr(dep, 'value', None), str) and dep.value.startswith('$'):

            # For security purposes, importing and instantiation symbols
            # from environment variables is not allowed.
            if dep.is_symbol():
                raise ConfigurationError("Can not use environment variable %s as symbol."
                    % dep.value)

            varname = dep.value[1:]
            value = os.getenv(varname)
            if value is None:
                raise EnvironmentVariableNotDefined(varname)
            dep.value = value

        # For literal dependencies, we only have to return the value
        # specified in the configuration file.
        if dep.is_literal():
            return dep.value

        # Symbol dependencies can be resolved using the import_symbol()
        # function.
        if dep.is_symbol():
            symbol = import_symbol(dep.value)
            if dep.requires_invocation():
                symbol = symbol(*dep.args, **dep.kwargs)

            #if dep.mode == 'append':
            #    seq = provider.provider._Provider__resolved.get(dep.name)
            #    seq.append(symbol)

            return symbol

        # Dependency collections are simple lists of dependencies.
        if dep.is_collection():
            return DependencyCollection(provider, dep.members)

        # In this implementation, nested dependencies MUST be declared
        # in order e.g. if B depends on A, then A has to be already
        # resolved. In practice this means that in the configuration file,
        # nested dependencies can not refer to dependencies declared after
        # themselves.
        for name in dep.requires():
            if not provider.is_satisfied(name):
                raise UnsatisfiedDependency(name)

        # If the dependency is in injected but does not have a factor,
        # it simply points to an already resolved declaration.
        if dep.is_injected() and not dep.has_factory():
            return provider.resolve(dep.key)

        # Get the factory, positional and keyword arguments, and
        # chained dependencies from the provider.
        f = dep.factory.resolve(self)
        args = [x.resolve(self) for x in dep.args]
        kwargs = {x: y.resolve(self) for x, y in dep.kwargs.items()}
        chain = [x.resolve(self) for x in dep.chain]

        # So now we invoke f using args and kwargs, and map it
        # to all functions in chain.
        result = f(*args, **kwargs)
        for f in chain:
            result = f(result)
        return result
